
// Waveblend - complex dualtree based image fusion
// (C) Copyright 2004 -- Sebastian Nowozin <nowozin@cs.tu-berlin.de>
//
// This file is part of Waveblend.
//
// Waveblend is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published
// by the Free Software Foundation; version 2 of the License.
//
// Waveblend is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The license is included with the distribution in the file 'LICENSE'.
//

/* WaveletProcessing.cs
 *
 * A lot of this code is based on the excellent Matlab code by
 * Shihua Cai, Keyong Li, Farras Abdelnour and Professor Ivan Selesnick, which
 * is available at
 * http://taco.poly.edu/WaveletSoftware/index.html
 *
 * Thank you very much for this work, without which this program would not
 * have been possible.
 */

using System;

public class
WaveletProcessing : ICloneable
{
	public object Clone ()
	{
		return (new WaveletProcessing ());
	}

	public void Convolve1d (IWavelet wave, ref double[] inVal,
		ref double[] outTrend, ref double[] outFluctuation)
	{
		SimpleArray saIn = new SimpleArray (inVal);
		SimpleArray saTrend = new SimpleArray (outTrend);
		SimpleArray saFluctuation = new SimpleArray (outFluctuation);

		Convolve (wave, saIn, saTrend, saFluctuation);
	}

	private void ConvolveAFBWorking (IWavelet wave, IWaveletArray inVal,
		IWaveletArray outTrend, IWaveletArray outFluctuation)
	{
		double[] loFilt = wave.ScalingNumbersAnalysis;
		double[] hiFilt = wave.WaveletNumbersAnalysis;
		int valClen = loFilt.Length + inVal.Length - 1;

		double[] trend = new double[valClen];
		double[] fluct = new double[valClen];

		int N = inVal.Length;
		int L = loFilt.Length / 2;

		for (int k = 0 ; k < valClen ; ++k) {
			for (int j = 0 ; j < loFilt.Length ; ++j) {
				int vIdx = k - j;

				if (vIdx < 0 || vIdx >= inVal.Length)
					continue;

				trend[k] += loFilt[j] * inVal[vIdx];
				fluct[k] += hiFilt[j] * inVal[vIdx];
			}
		}

		// Debug output
		foreach (double val in trend)
			Console.WriteLine ("trend: {0}", val);
		Console.WriteLine ("");
		foreach (double val in fluct)
			Console.WriteLine ("fluct: {0}", val);
	}

	// Convolution
	//
	// Preconditions:
	// inVal.Length % 2 == 0
	// wave.SupportSize >= inVal.Length
	// outTrend.Length == outFluctuation.Length == inVal.Length / 2
	//
	private void Convolve (IWavelet wave, IWaveletArray inVal,
		IWaveletArray outTrend, IWaveletArray outFluctuation)
	{
		/*Console.WriteLine ("Convolve (\"{0}\", inVal[{1}], ..)",
			wave.Name, inVal.Length);*/

		double[] loFilt = wave.ScalingNumbersAnalysis;
		double[] hiFilt = wave.WaveletNumbersAnalysis;

		ConvolveArr (loFilt, hiFilt, inVal, outTrend, outFluctuation);
	}

	private void ConvolveArr (double[] loFilt, double[] hiFilt,
		IWaveletArray inVal,
		IWaveletArray outTrend, IWaveletArray outFluctuation)
	{
		int valClen = loFilt.Length + inVal.Length - 1;

		int N = inVal.Length;
		int L = loFilt.Length / 2;
		int ON = N / 2;

		for (int k = 0 ; k < ((valClen + 1) / 2) ; ++k) {
			for (int j = 0 ; j < loFilt.Length ; ++j) {
				int vIdx = 2*k - j;

				if (vIdx < 0 || vIdx >= inVal.Length)
					continue;

				vIdx += L;
				vIdx %= inVal.Length;

				if (k >= N)
					continue;

				outTrend[k % ON] += loFilt[j] * inVal[vIdx];
				outFluctuation[k % ON] += hiFilt[j] * inVal[vIdx];
			}
		}
	}


	public void Convolve2d (IWavelet wave, double[,] inVal,
		double[,] outTrend, double[,] outH,
		double[,] outD, double[,] outV)
	{
		Convolve2dArr (wave.ScalingNumbersAnalysis, wave.WaveletNumbersAnalysis,
			wave.ScalingNumbersAnalysis, wave.WaveletNumbersAnalysis,
			inVal, outTrend, outH, outD, outV);
	}

	public void Convolve2dArr (double[] loFiltCol, double[] hiFiltCol,
		double[] loFiltRow, double[] hiFiltRow,
		double[,] inVal,
		double[,] outTrend, double[,] outH,
		double[,] outD, double[,] outV)
	{
		if ((inVal.GetLength (0) % 2) != 0 ||
			(inVal.GetLength (1) % 2) != 0)
		{
			throw (new ArgumentException (String.Format
				("Input array has to be of even dimensions, got ({0}x{1}).",
				inVal.GetLength (0), inVal.GetLength (1))));
		}

		double[,] rowTrends =
			new double[inVal.GetLength (0), inVal.GetLength (1) / 2];
		double[,] rowFlucts =
			new double[inVal.GetLength (0), inVal.GetLength (1) / 2];

		// From F create T_{Row} and F_{Row}
		for (int row = 0 ; row < inVal.GetLength (0) ; ++row) {
			ConvolveArr (loFiltCol, hiFiltCol,
				new DirectionalArray (DirectionalArray.Direction.Row, inVal, row),
				new DirectionalArray (DirectionalArray.Direction.Row, rowTrends, row),
				new DirectionalArray (DirectionalArray.Direction.Row, rowFlucts, row));
		}

		// From T_{Row} create H and A (Trend)
		for (int col = 0 ; col < rowTrends.GetLength (1) ; ++col) {
			ConvolveArr (loFiltRow, hiFiltRow,
				new DirectionalArray (DirectionalArray.Direction.Column, rowTrends, col),
				new DirectionalArray (DirectionalArray.Direction.Column, outTrend, col),
				new DirectionalArray (DirectionalArray.Direction.Column, outH, col));
		}

		// From F_{Row} create V and D
		for (int col = 0 ; col < rowFlucts.GetLength (1) ; ++col) {
			ConvolveArr (loFiltRow, hiFiltRow,
				new DirectionalArray (DirectionalArray.Direction.Column, rowFlucts, col),
				new DirectionalArray (DirectionalArray.Direction.Column, outV, col),
				new DirectionalArray (DirectionalArray.Direction.Column, outD, col));
		}
	}

	//
	// 2D DT Complex Discrete Wavelet Transform
	//

	// Returns two trees, each which hold 3 complex subbands per level plus
	// the trend signal. That are six real signals per level.
	//
	// The returned array 'dt' has the following structure:
	//
	// dt[0] holds the first 3 subband directions
	// dt[1] holds the second 3 subband directions
	//
	// dt[0][0-2] are the three the ComplexArray H, D, V bands of plane 1
	// dt[1][0-2] are the other three the ComplexArray H, D, V bands of
	//   plane 2
	// dt[0].GetTrend (0/1) gets the first or second trend (lowpass subband)
	//   signals of plane 1
	// dt[1].GetTrend (0/1) gets the first or second trend (lowpass subband)
	//   signals of plane 2

	public Dualtree2dComplex[] DualtreeTransform2dComplex (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, double[,] inVal, int scale)
	{
		Dualtree2dComplex[] trees = new Dualtree2dComplex[2];
		Dualtree2dReal[] treesReal = new Dualtree2dReal[2];

		// Orientation index t2
		for (int t2 = 0 ; t2 < 2 ; ++t2) {
			// Real/imaginary index t1
			for (int t1 = 0 ; t1 < 2 ; ++t1) {
				/*Console.WriteLine ("DualtreeTansform2dComplex: creating real tree ({0}, {1})",
					t1 == 0 ? "Real" : "Imaginary", t2 == 0 ? "Orientation 1" : "Orientation 2");*/

				treesReal[t1] = DualtreeTransform2dRealSingleTree (wave1rst,
					wave, inVal, scale, t1, t2);
			}

			trees[t2] = new Dualtree2dComplex (treesReal[0], treesReal[1]);
		}

		// Now trees[0] is 1 orientation, trees[1] is 2 orientation, each
		// holding 3 subbands.

		Dualtree2dComplex tree1 = trees[0];
		Dualtree2dComplex tree2 = trees[1];

		while (tree1 != null) {
			for (int m = 0 ; m <= 2 ; ++m) {
				DualtreeTransform2dComplexSubtract (tree1[m], 0, tree2[m], 1);
				DualtreeTransform2dComplexSubtract (tree2[m], 0, tree1[m], 1);
			}

			tree1 = tree1.Next;
			tree2 = tree2.Next;
		}

		return (trees);
	}

	private void DualtreeTransform2dComplexSubtract
		(Dualtree2dComplex.ComplexArray c1, int ri1,
		Dualtree2dComplex.ComplexArray c2, int ri2)
	{
		/*Console.WriteLine ("DualtreeTransform2dComplexSubtract (c1, {0}, c2, {1})",
			ri1, ri2);*/
		double sqrt2 = Math.Sqrt (2.0);

		for (int y = 0 ; y < c1.GetLength (0) ; ++y) {
			for (int x = 0 ; x < c1.GetLength (1) ; ++x) {
				double u = (c1[y, x, ri1] + c2[y, x, ri2]) / sqrt2;

				c2[y, x, ri2] = (c1[y, x, ri1] - c2[y, x, ri2]) / sqrt2;
				c1[y, x, ri1] = u;
			}
		}
	}

	// Complex 2d DT INVERSION
	// Warning: this changes the original data stored in the tree.
	//
	public double[,] DualtreeTransform2dComplexInverse (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, Dualtree2dComplex[] trees)
	{
		Dualtree2dComplex tree1 = trees[0];
		Dualtree2dComplex tree2 = trees[1];

		while (tree1 != null) {
			for (int m = 0 ; m <= 2 ; ++m) {
				DualtreeTransform2dComplexSubtract (tree1[m], 0, tree2[m], 1);
				DualtreeTransform2dComplexSubtract (tree2[m], 0, tree1[m], 1);
			}

			tree1 = tree1.Next;
			tree2 = tree2.Next;
		}

		// The magic output array.
		double[,] result = new double[trees[0].GetTrend (0).GetLength (0) * 2,
			trees[0].GetTrend (0).GetLength (1) * 2];

		// Orientation index t2 (tree index)
		for (int t2 = 0 ; t2 < 2 ; ++t2) {
			// Real/imaginary index t1
			for (int t1 = 0 ; t1 < 2 ; ++t1) {
				double[,] trend = DualtreeTransform2dRealInverseSingleTree
					(wave1rst, wave, trees[t2].ProduceRealDualtree (t1),
					t1, t2, true);

				for (int y = 0 ; y < trend.GetLength (0) ; ++y)
					for (int x = 0 ; x < trend.GetLength (1) ; ++x)
						result[y, x] += trend[y, x];
			}
		}

		for (int y = 0 ; y < result.GetLength (0) ; ++y)
			for (int x = 0 ; x < result.GetLength (1) ; ++x)
				result[y, x] /= 4.0;

		return (result);
	}


	//
	// 2D DT Real Discrete Wavelet Transform
	//

	public Dualtree2dReal[] DualtreeTransform2dReal (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, double[,] inVal, int scale)
	{
		Dualtree2dReal[] trees = new Dualtree2dReal[2];

		for (int t = 0 ; t < 2 ; ++t)
			trees[t] = DualtreeTransform2dRealSingleTree (wave1rst,
				wave, inVal, scale, t);

		// Do the final add/subtraction and normalization step for each
		// fluctuation subsignal
		Dualtree2dReal tree1 = trees[0];
		Dualtree2dReal tree2 = trees[1];

		while (tree1 != null) {
			for (int m = 1 ; m <= 3 ; ++m)
				DualtreeTransform2dSubtract ((double[,]) tree1[m],
					(double[,]) tree2[m]);

			tree1 = tree1.Next;
			tree2 = tree2.Next;
		}

		return (trees);
	}

	private void DualtreeTransform2dSubtract (double[,] s1,
		double[,] s2)
	{
		double sqrt2 = Math.Sqrt (2.0);

		for (int y = 0 ; y < s1.GetLength (0) ; ++y) {
			for (int x = 0 ; x < s1.GetLength (1) ; ++x) {
				double u = (s1[y,x] + s2[y,x]) / sqrt2;

				s2[y,x] = (s1[y,x] - s2[y,x]) / sqrt2;
				s1[y,x] = u;
			}
		}
	}

	private Dualtree2dReal DualtreeTransform2dRealSingleTree (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, double[,] inVal, int scale, int tree)
	{
		return (DualtreeTransform2dRealSingleTree (wave1rst, wave, inVal,
			scale, tree, tree));
	}

	private Dualtree2dReal DualtreeTransform2dRealSingleTree (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, double[,] inVal, int scale, int t1, int t2)
	{
		Dualtree2dReal root = DualtreeTransform2dRealSub (wave1rst, t1, t2, inVal);
		Dualtree2dReal walk = root;

		for (int k = 1 ; k < scale ; ++k) {
			double[,] trend = (double[,]) walk[0];

			Dualtree2dReal subtree = DualtreeTransform2dRealSub (wave, t1, t2,
				trend);

			walk.Next = subtree;
			walk = subtree;
		}
		walk.LastLevel = true;

		return (root);
	}

	private Dualtree2dReal DualtreeTransform2dRealSub (IWaveletDTCW wave,
		int t1, int t2, double[,] inVal)
	{
		Dualtree2dReal dt = new Dualtree2dReal ();

		int yDim = inVal.GetLength (0) / 2;
		int xDim = inVal.GetLength (1) / 2;

		double[,] trend = new double[yDim, xDim];
		double[,] H = new double[yDim, xDim];
		double[,] D = new double[yDim, xDim];
		double[,] V = new double[yDim, xDim];

		Convolve2dArr (wave[FilterType.Analysis, t1, PassType.Trend],
			wave[FilterType.Analysis, t1, PassType.Fluctuation],
			wave[FilterType.Analysis, t2, PassType.Trend],
			wave[FilterType.Analysis, t2, PassType.Fluctuation],
			inVal, trend, H, D, V);

		dt[0] = trend;
		dt[2] = H;
		dt[3] = D;
		dt[1] = V;

		return (dt);
	}

	// Warning: this changes the original data stored in the tree.
	public double[,] DualtreeTransform2dRealInverse (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, Dualtree2dReal[] trees)
	{
		// Sum and difference (why?)
		Dualtree2dReal tree1 = trees[0];
		Dualtree2dReal tree2 = trees[1];

		while (tree1 != null) {
			for (int m = 1 ; m <= 3 ; ++m)
				DualtreeTransform2dSubtract ((double[,]) tree1[m],
					(double[,]) tree2[m]);

			tree1 = tree1.Next;
			tree2 = tree2.Next;
		}

		double[][,] outTrends = new double[2][,];

		for (int n = 0 ; n < 2 ; ++n)
			outTrends[n] = DualtreeTransform2dRealInverseSingleTree
				(wave1rst, wave, trees[n], n);

		double[,] outData = new double[outTrends[0].GetLength (0),
			outTrends[0].GetLength (1)];

		for (int y = 0 ; y < outTrends[0].GetLength (0) ; ++y) {
			for (int x = 0 ; x < outTrends[0].GetLength (1) ; ++x) {
				outData[y, x] = (outTrends[0][y, x] +
					outTrends[1][y, x]) / 2.0; // FIXME: Farras uses sqrt(2.0) here, why?
			}
		}

		return (outData);
	}

	private double[,] DualtreeTransform2dRealInverseSingleTree
		(IWaveletDTCW wave1rst, IWaveletDTCW wave, Dualtree2dReal root,
		int tree)
	{
		return (DualtreeTransform2dRealInverseSingleTree (wave1rst, wave,
			root, tree, true));
	}

	private double[,] DualtreeTransform2dRealInverseSingleTree
		(IWaveletDTCW wave1rst, IWaveletDTCW wave, Dualtree2dReal root,
		int tree, bool first)
	{
		return (DualtreeTransform2dRealInverseSingleTree (wave1rst, wave,
			root, tree, tree, first));
	}

	private double[,] DualtreeTransform2dRealInverseSingleTree
		(IWaveletDTCW wave1rst, IWaveletDTCW wave, Dualtree2dReal root,
		int t1, int t2, bool first)
	{
		/*Console.WriteLine ("DualtreeTransform2dRealInverseSingleTree (tree {0}|{1}, first {2})",
			t1, t2, first);*/

		double[,] subTrend;

		if (root.LastLevel) {
			subTrend = (double[,]) root[0];
		} else {
			subTrend = DualtreeTransform2dRealInverseSingleTree (wave1rst,
				wave, root.Next, t1, t2, false);
		}

		double[,] H = (double[,]) root[2];
		double[,] D = (double[,]) root[3];
		double[,] V = (double[,]) root[1];

		double[,] upTrend = DualtreeTransform2dRealInverseSub (subTrend, H, D, V,
			first ? wave1rst : wave, t1, t2);

		return (upTrend);
	}

	private double[,] DualtreeTransform2dRealInverseSub (double[,] trend,
		double[,] H, double[,] D, double[,] V,
		IWaveletDTCW wave, int t1, int t2)
	{
		double[,] outVal = new double[trend.GetLength (0) * 2,
			trend.GetLength (1) * 2];

		/*Console.WriteLine ("DualtreeTransform2dRealInverseSub: to {0}x{1}",
			outVal.GetLength (0), outVal.GetLength (1));*/

		Inverse2dArr (wave[FilterType.Synthesis, t1, PassType.Trend],
			wave[FilterType.Synthesis, t1, PassType.Fluctuation],
			wave[FilterType.Synthesis, t2, PassType.Trend],
			wave[FilterType.Synthesis, t2, PassType.Fluctuation],
			outVal, trend, H, D, V);

		return (outVal);
	}

	//
	// 1D DT Real Discrete Wavelet Transform
	//

	public Dualtree[] DualtreeTransform1d (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, double[] inVal, int scale)
	{
		SimpleArray saIn = new SimpleArray (inVal);

		Dualtree[] trees = new Dualtree[2];

		for (int n = 0 ; n < 2 ; ++n)
			trees[n] = DualtreeTransform1dSingleTree (wave1rst,
				wave, saIn, scale, n);

		return (trees);
	}

	private Dualtree DualtreeTransform1dSingleTree (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, SimpleArray saIn, int scale, int tree)
	{
		// Build tree, starting from root node
		Dualtree root = DualtreeTransform (wave1rst, tree, saIn);
		Dualtree walk = root;

		for (int k = 1 ; k < scale ; ++k) {
			double[] trend = (double[]) walk[0];

			Dualtree subtree = DualtreeTransform (wave, tree,
				new SimpleArray (trend));

			walk.Next = subtree;
			walk = subtree;
		}
		walk.LastLevel = true;

		return (root);
	}

	private Dualtree DualtreeTransform (IWaveletDTCW wave, int tree,
		IWaveletArray inVal)
	{
		Dualtree dt = new Dualtree ();

		double[] trend = new double[inVal.Length / 2];
		double[] fluctuation = new double[inVal.Length / 2];

		ConvolveArr (wave[FilterType.Analysis, tree, PassType.Trend],
			wave[FilterType.Analysis, tree, PassType.Fluctuation],
			inVal, new SimpleArray (trend), new SimpleArray (fluctuation));

		dt[0] = trend;
		dt[1] = fluctuation;

		return (dt);
	}


	public double[] DualtreeTransform1dInverse (IWaveletDTCW wave1rst,
		IWaveletDTCW wave, Dualtree[] trees)
	{
		double[][] outTrends = new double[2][];

		for (int n = 0 ; n < 2 ; ++n)
			outTrends[n] = DualtreeTransform1dInverseSingleTree
				(wave1rst, wave, trees[n], n, true);

		double[] outData = new double[outTrends[0].Length];
		for (int n = 0 ; n < outTrends[0].Length ; ++n)
			outData[n] = (outTrends[0][n] + outTrends[1][n]) / 2.0; //Math.Sqrt (2.0);

		return (outData);
	}

	private double[] DualtreeTransform1dInverseSingleTree
		(IWaveletDTCW wave1rst, IWaveletDTCW wave, Dualtree root,
		int tree, bool first)
	{
		double[] subTrend;
		double[] subFluctuation = (double[]) root[1];

		if (root.LastLevel) {
			subTrend = (double[]) root[0];
		} else {
			subTrend = DualtreeTransform1dInverseSingleTree (wave1rst, wave,
				root.Next, tree, false);
		}

		double[] upTrend = DualtreeTransformInverse (subTrend,
			subFluctuation, first ? wave1rst : wave, tree);

		return (upTrend);
	}

	// Inverses the trend and fluctuation using the wavelet 'wave' to yield
	// the upsampled trend signal as output. 'tree' denotes if its tree 0 or
	// tree 1.
	private double[] DualtreeTransformInverse (double[] trend,
		double[] fluctuation, IWaveletDTCW wave, int tree)
	{
		double[] outVal = new double[trend.Length * 2];

		InverseArr (wave[FilterType.Synthesis, tree, PassType.Trend],
			wave[FilterType.Synthesis, tree, PassType.Fluctuation],
			new SimpleArray (outVal),
			new SimpleArray (trend), new SimpleArray (fluctuation));

		return (outVal);
	}

	public double[,] tileCombine2d (double[,] trend, double[,] H,
		double[,] D, double[,] V, bool normalize)
	{
		int x = trend.GetLength (0);
		int y = trend.GetLength (1);
		double[,] res = new double[2 * x, 2 * y];

		if (normalize) {
			ImageMap tMap = new ImageMap (trend);
			tMap.Normalize ();
			trend = tMap.ValueArray;

			tMap = new ImageMap (H);
			tMap.Normalize ();
			H = tMap.ValueArray;

			tMap = new ImageMap (D);
			tMap.Normalize ();
			D = tMap.ValueArray;

			tMap = new ImageMap (V);
			tMap.Normalize ();
			V = tMap.ValueArray;
		}

		placeCopy (ref res, 0, y, trend);
		placeCopy (ref res, 0, 0, H);
		placeCopy (ref res, x, 0, D);
		placeCopy (ref res, x, y, V);

		return (res);
	}

	private void placeCopy (ref double[,] res, int wx, int wy, double[,] source)
	{
		for (int y = 0 ; y < source.GetLength (0) ; ++y)
			for (int x = 0 ; x < source.GetLength (1) ; ++x)
				res[y + wy, x + wx] = source[y, x];
	}

	public void Inverse1d (IWavelet wave, double[] function,
		double[] trend, double[] fluctuation)
	{
		SimpleArray saOut = new SimpleArray (function);
		SimpleArray saTrend = new SimpleArray (trend);
		SimpleArray saFluctuation = new SimpleArray (fluctuation);

		Inverse (wave, saOut, saTrend, saFluctuation);
	}

	private void InverseSFBWorking (IWavelet wave, IWaveletArray function,
		IWaveletArray trend, IWaveletArray fluctuation)
	{
		/*Console.WriteLine ("InverseSFB (\"{0}\", function[{1}], ..)",
			wave.Name, function.Length);*/

		double[] loFilt = wave.ScalingNumbersSynthesis;
		double[] hiFilt = wave.WaveletNumbersSynthesis;

		// Upsample
		double[] trend0 = new double[2 * trend.Length];
		double[] fluct0 = new double[2 * fluctuation.Length];
		for (int n = 0 ; n < trend.Length ; ++n) {
			trend0[2 * n] = trend[n];
			fluct0[2 * n] = fluctuation[n];
		}


		int N = 2 * trend.Length;
		int L = loFilt.Length;

		int valClen = loFilt.Length + trend0.Length - 1;
		double[] outValHi = new double[valClen];
		double[] outValLo = new double[valClen];

		for (int k = 0 ; k < valClen ; ++k) {
			for (int j = 0 ; j < loFilt.Length ; ++j) {
				int vIdx = k - j;

				//Console.WriteLine ("vIdx = {0}, k = {1}", vIdx, k);
				if (vIdx < 0 || vIdx >= trend0.Length)
					continue;

				outValLo[k] += loFilt[j] * trend0[vIdx];
				outValHi[k] += hiFilt[j] * fluct0[vIdx];
				//outFluctuation[k % ON] += hiFilt[j] * inVal[vIdx];
			}
		}

		double[] outVal = new double[valClen];
		for (int n = 0 ; n < valClen ; ++n)
			outVal[n] = outValHi[n] + outValLo[n];

		for (int n = 0 ; n < (L - 2) ; ++n)
			outVal[n] += outVal[n + N];

		int shift = L / 2 - 1;
		for (int n = 0 ; n < N ; ++n)
			function[n] = outVal[(n + shift) % N];

		/*
		foreach (double val in outVal)
			Console.WriteLine ("out: {0}", val);
			*/
	}

	private void Inverse (IWavelet wave, IWaveletArray function,
		IWaveletArray trend, IWaveletArray fluctuation)
	{
		/*Console.WriteLine ("InverseSFB (\"{0}\", function[{1}], ..)",
			wave.Name, function.Length);*/

		double[] loFilt = wave.ScalingNumbersSynthesis;
		double[] hiFilt = wave.WaveletNumbersSynthesis;

		InverseArr (loFilt, hiFilt, function, trend, fluctuation);
	}

	private void InverseArr (double[] loFilt, double[] hiFilt,
		IWaveletArray function,
		IWaveletArray trend, IWaveletArray fluctuation)
	{
		int N = 2 * trend.Length;
		int L = loFilt.Length;

		int valClen = loFilt.Length + N - 1;
		double[] outVal = new double[N];

		// Reset
		for (int n = 0 ; n < N ; ++n)
			function[n] = 0.0;

		// Circular shift in final array assignment
		int shift = -(L / 2 - 1);
		if (shift < 0)
			shift += N;

		for (int k = 0 ; k < valClen ; ++k) {
			for (int j = 0 ; j < loFilt.Length ; ++j) {
				int vIdx = k - j;

				if (vIdx < 0 || vIdx >= N)
					continue;

				if (vIdx % 2 == 1)
					continue;

				double tv = trend[vIdx / 2];
				double fv = fluctuation[vIdx / 2];

				int dk = k;
				if (k >= N && k < (N + L - 2))
					dk -= N;
				if (dk >= N)
					continue;

				function[(dk + shift) % N] += loFilt[j] * tv + hiFilt[j] * fv;
			}
		}
	}

	// Reconstructs the upsampled signal in function from the four subsignals
	// trend, H, D, V.
	public void Inverse2d (IWavelet wave, double[,] function,
		double[,] trend, double[,] H,
		double[,] D, double[,] V)
	{
		Inverse2dArr (wave.ScalingNumbersAnalysis, wave.WaveletNumbersAnalysis,
			wave.ScalingNumbersAnalysis, wave.WaveletNumbersAnalysis,
			function, trend, H, D, V);
	}

	public void Inverse2dArr (double[] loFiltCol, double[] hiFiltCol,
		double[] loFiltRow, double[] hiFiltRow,
		double[,] function,
		double[,] trend, double[,] H,
		double[,] D, double[,] V)
	{
		if ((function.GetLength (0) % 2) != 0 ||
			(function.GetLength (1) % 2) != 0)
		{
			throw (new ArgumentException ("Output array has to be of even dimensions."));
		}

		double[,] rowTrends =
			new double[function.GetLength (0), function.GetLength (1) / 2];
		double[,] rowFlucts =
			new double[function.GetLength (0), function.GetLength (1) / 2];


		// Create F_{Row} from V and D
		for (int col = 0 ; col < rowFlucts.GetLength (1) ; ++col) {
			InverseArr (loFiltRow, hiFiltRow,
				new DirectionalArray (DirectionalArray.Direction.Column, rowFlucts, col),
				new DirectionalArray (DirectionalArray.Direction.Column, V, col),
				new DirectionalArray (DirectionalArray.Direction.Column, D, col));
		}

		// Create T_{Row} from H and A (Trend)
		for (int col = 0 ; col < rowTrends.GetLength (1) ; ++col) {
			InverseArr (loFiltRow, hiFiltRow,
				new DirectionalArray (DirectionalArray.Direction.Column, rowTrends, col),
				new DirectionalArray (DirectionalArray.Direction.Column, trend, col),
				new DirectionalArray (DirectionalArray.Direction.Column, H, col));
		}

		// Create F from T_{Row} and F_{Row}
		for (int row = 0 ; row < function.GetLength (0) ; ++row) {
			InverseArr (loFiltCol, hiFiltCol,
				new DirectionalArray (DirectionalArray.Direction.Row, function, row),
				new DirectionalArray (DirectionalArray.Direction.Row, rowTrends, row),
				new DirectionalArray (DirectionalArray.Direction.Row, rowFlucts, row));
		}
	}


	// Wrapper classes so we can share all of the wavelet code between 1d and
	// 2d wavelet operations by mapping 2d arrays to 1d array slices.
	private interface
	IWaveletArray
	{
		int Length {
			get;
		}

		double this[int n] {
			get;
			set;
		}
	}

	private class
	SimpleArray : IWaveletArray
	{
		private SimpleArray ()
		{
		}

		double[] arr;

		public SimpleArray (double[] arr)
		{
			this.arr = arr;
		}

		public int Length {
			get {
				return (arr.Length);
			}
		}

		public double this[int n] {
			get {
				return (arr[n]);
			}
			set {
				arr[n] = value;
			}
		}
	}

	private class
	DirectionalArray : IWaveletArray
	{
		private DirectionalArray ()
		{
		}

		public enum Direction {
			Row,
			Column,
		};

		Direction dir;
		double[,] arr;
		int resolveDim;

		public DirectionalArray (Direction dir, double[,] arr, int resolveDim)
		{
			this.dir = dir;
			this.arr = arr;
			this.resolveDim = resolveDim;
		}

		public int Length {
			get {
				if (dir == Direction.Row)
					return (arr.GetLength (1));
				else
					return (arr.GetLength (0));
			}
		}

		public double this[int n] {
			get {
				/*Console.WriteLine ("get: {0} of {1}, dimension {2}",
					n, (dir == Direction.Row) ? arr.GetLength (1) : arr.GetLength (0),
					dir);*/

				if (dir == Direction.Row)
					return (arr[resolveDim, n]);
				else
					return (arr[n, resolveDim]);
			}

			set {
				if (dir == Direction.Row)
					arr[resolveDim, n] = value;
				else
					arr[n, resolveDim] = value;
			}
		}
	}
}


